# Model Configuration

The model configuration is specified within the `PlanktonModel` data structure. As documented below, the various parameters of a `PlanktonModel` can all be specified via keyword arguments. Options and features are documented in more detail afterwards and in the [Examples](@ref examples) section.

```@docs
PlanktonModel
```

## Architecture

Passing `arch = CPU()` or `arch = GPU()` to the `PlanktonModel` constructor will determine whether the model
is time stepped on a CPU or GPU.

Users do not need to modify the setup or simulation script to change the architecture to run on.
The only thing that needs to be changed is `arch = CPU()` or `arch = GPU()`.

!!! tip "Running on GPUs"
    Please refer to [GPU Support](@ref) for more detail on running `PlanktonIndividuals` on GPUs and don't hesitate to [open an issue](https://github.com/JuliaOcean/PlanktonIndividuals.jl/issues/new) if you have any difficulty.

## Grid

Two options are supported for the grid: `RegularRectilinearGrid` and `RegularLatLonGrid` with constant grid spacing in each
direction; grid spacing can differ between dimensions. Both `Periodic` and `Bounded` domain options are supported for the horizontal directions, whereas the domain is always `Bounded` in the vertical direction (top and bottom).

The `RegularRectilinearGrid` is constructed by specifying its `size` (`Tuple`
specifying the number of grid points in each dimension) and `spacing` (`Tuple` specifying
the grid cell width for each dimension).

For example, a rectilinear grid with ``32 \times 64 \times 128`` grid points and grid `spacing` of ``dx=1``m, ``dy=2``m, ``dz=4``m is constructed like this:

```@setup 1
using PlanktonIndividuals
```

```@repl 1
grid = RegularRectilinearGrid(size=(32, 64, 128), spacing=(1.0, 2.0, 4.0))
```

The `RegularLatLonGrid` is constructed by specifying its `size` (`Tuple`
specifying the number of grid points in each dimension) and `lat`, `lon`, `z` (`Tuple` specifying
the start and end points).

For example, a global domain from 80S to 80N, 180W to 180E and 200m depth with spacing of 1 degree horizontally and 10m vertically
is constructed like this:

```@repl 1
grid = RegularLatLonGrid(size=(360,160,20), lat = (-80,80), lon = (-180,180), z = (0,-200)) 
```

## Individuals

The number of species can be specified via `N_species`.
The number of individuals per species can be specified via `N_individual`.
The maximum number of individuals per species that the model can hold is specified via `max_individuals`.

## [Parameters](@id parameters)

Default parameters are generated by two functions called [`bgc_params_default`](@ref) and [`phyt_params_default`](@ref) via dictionnaries (`Dict`).
`bgc_params` contains the parameters for biogeochemical cycls.
`phyt_params` contains the parameters for phytoplankton individuals.
The default parameter values are listed [here](https://github.com/JuliaOcean/PlanktonIndividuals.jl/blob/master/src/params/param_default.jl).

Values of the parameters can be changed using [`update_bgc_params`](@ref) and [`update_phyt_params`](@ref).

In the example shown below, we change the value of `kDOC`, which is the remineralization rate for `DOC`:

```@repl 1
new_params = Dict("kDOC" => 0.01) # no need to include all parameters
update_bgc_params(new_params)
```

`phyt_params` can be changed in the same way.

## Nutrient fields

Nutrient fields included in the model are listed below.
The initial conditions of the nutrient fields are generated by [`generate_nutrients`](@ref)
either from a `NamedTuple` or a `Dict`. The `NamedTuple` contains two elements. First element is a 10-element
`NamedTuple`, each element is a `Float64` number filled uniformly across the domain as initial condition of a
tracer. Second element is a 10-element `NamedTuple`, each element is a `Float64` number indicating the random noise of each tracer. The `Dict` contains the file paths pointing to the files of existing nutrient initial conditions.

|Name | Unit        | Description                                     |
|-----|-------------|-------------------------------------------------|
|DIC  |$mmol~C/m^3$ | concentration of dissolved inorganic carbon     |
|NH4  |$mmol~N/m^3$ | concentration of ammonia                        |
|NO3  |$mmol~N/m^3$ | concentration of nitrate                        |
|PO4  |$mmol~P/m^3$ | concentration of phosphate                      |
|DOC  |$mmol~C/m^3$ | concentration of dissolved organic carbon       |
|DON  |$mmol~N/m^3$ | concentration of dissolved organic nitrogen     |
|DOP  |$mmol~P/m^3$ | concentration of dissolved organic phosphorus   |
|POC  |$mmol~C/m^3$ | concentration of particulate organic carbon     |
|PON  |$mmol~N/m^3$ | concentration of particulate organic nitrogen   |
|POP  |$mmol~P/m^3$ | concentration of particulate organic phosphorus |

!!! tip "Noisy Initial Conditions"
    A random noise can be included only if the initial conditions are generated by `NamedTuple`.

!!! tip "Nutrient Fields"
    The initial conditions of all the nutrient fields should be non-negative.

An example of the `NamedTuple` is listed below:

```@repl 1
initial_condition = (DIC=20.0, NH4=0.2, NO3=0.5, PO4=0.03, DOC=1.0, DON=0.1, DOP=0.05, POC=0.0, PON=0.0,POP=0.0);
rand_noise = (DIC=0.0, NH4=0.0, NO3=0.0, PO4=0.0, DOC=0.0, DON=0.0, DOP=0.0, POC=0.0, PON=0.0,POP=0.0);
nut_initial = (initial_condition = initial_condition, rand_noise = rand_noise)
```

And example of the `Dict` is listed below:

```@repl 1
nut_init = Dict(
    "DIC" => "path/to/DIC.bin",
    "NH4" => "path/to/NH4.bin",
    "NO3" => "path/to/NO3.bin",
    "PO4" => "path/to/PO4.bin",
    "DOC" => "path/to/DOC.bin",
    "DON" => "path/to/DON.bin",
    "DOP" => "path/to/DOP.bin",
    "POC" => "path/to/POC.bin",
    "PON" => "path/to/PON.bin",
    "POP" => "path/to/POP.bin");
```
